--  Copyright  (C) 2006 M. Ben-Ari. See copyright.txt

with Semaphore_Package;
with Ada.Text_IO;
with Ada.Finalization;
with Ada.Exceptions;

procedure CL is
   --  the program main
   --  named after Chandy & Lamport
   use Semaphore_Package;

   --  the system has Node_Count'Last nodes
   --  the value 0 denotes no neighbour
   type Node_Count is range 0 .. 4;
   subtype Node_ID is Node_Count range 1 .. Node_Count'Last;

   Marker : constant Integer := -1; -- the marker

   Slow_Down : constant Duration := 0.3; -- arbitrary slow-down factor

   type Channel is --  for nodes to know the system topology
      record
         Exists         : Boolean  := False;
         Last_Message   : Integer  := 0;
         Recorded_State : Integer  := 0;
         Marker_State   : Integer  := Marker;
      end record;

   --  the whole span of work (can be any)
   subtype Workload is Positive range 1 .. 42;
   Stop_Point : constant Workload := 21;

   --  a mutual-exclusion semaphore to protect printing
   Print : constant Binary_Semaphore  := Init (1);

   --  a mutual-exclusion semaphore to protect state handling
   Sem : constant Binary_Semaphore  := Init (1);

   ------------------- to play with finalization
   type Last_Wishes_T is new Ada.Finalization.Limited_Controlled with
      record
         ID : Node_ID;
      end record;
   procedure Finalize (Last_Wishes : in out Last_Wishes_T) is
      use Ada.Text_IO;
   begin
      Put_Line
        (" Node" & Node_ID'Image (Last_Wishes.ID) &
         " gracefully terminated");
   end Finalize;
   ----------------------------------

   task type Node is
      ------------------------
      entry Init (ID  : Node_ID;
                  N_I : Node_Count;
                  N_O : Node_Count);
      entry Configure (C : Node_ID);
      ------------------------
      entry Message (M : Integer; ID : Node_ID);
   end Node;

   --  the system
   Nodes : array (Node_ID) of Node;
   ----------------------------------

   task body Node is

      Incoming : array (Node_ID) of Channel;
      pragma Volatile (Incoming);
      --  All tasks of the program that read or write volative variables
      --  see the same order of updates to those variables

      Outgoing : array (Node_ID) of Channel;
      pragma Volatile (Outgoing);

      State_Recorded : Boolean  := False;
      pragma Volatile (State_Recorded);

      N_In, N_Out   : Node_Count  := 0;
      Received_ID   : Node_ID;
      Received_Data : Integer;

      Current : Node_ID; --  to tell apart instances of core activity

      --------------- to play with finalization
      Last_Wishes : Last_Wishes_T; -- never read but causing finalization
      ----------------------------

      procedure Do_Work (Data : Integer) is
      begin
         for N in Node_ID loop
            if Outgoing (N).Exists then
               Wait (Sem); -- lock
               if Data >= 0 then
                  --  normal work: record latest outgoing message
                  Outgoing (N).Last_Message := Data;
               elsif not State_Recorded then
                  --  the marker has arrived,
                  --  but the state was not recorded already;
                  --  add the latest outgoing message to the state
                  Outgoing (N).Recorded_State  := Outgoing (N).Last_Message;
               end if;
               Signal (Sem); -- unlock
               Nodes (N).Message (Data, Current);
               delay Slow_Down;
            end if;
         end loop;
      end Do_Work;

      procedure Record_State is
      begin
         if not State_Recorded then
            --  state not being recorded yet
            Wait (Sem); --  lock
            --  record in state the latest incoming message
            --  for all incoming channles
            for N in Node_ID loop
               if Incoming (N).Exists then
                  Incoming (N).Recorded_State := Incoming (N).Last_Message;
               end if;
            end loop;
            Signal (Sem); --  unlock
            --  send the marker down all outgoing channels
            Do_Work (Marker);
            --  record this node is in "state recording" mode
            State_Recorded  := True;
         end if;
      end Record_State;

      function State_Written return Boolean is
      begin
         if not State_Recorded then
            return False;
         end if;
         --  we are here if we are in "state recording" mode
         for N in Node_ID loop
            if Incoming (N).Exists and Incoming (N).Marker_State < 0 then
               --  some incoming channel has sent no msg down yet:
               --  there may still be pending work
               return False;
            end if;
         end loop;
         --  else the state has been fully recorded for all channels
         Wait (Print); -- lock
         Ada.Text_IO.Put_Line (" Node" & Node_ID'Image (Current));
         Ada.Text_IO.Put_Line ("       Out channels");
         for N in Node_ID loop
            if Outgoing (N).Exists then
               if Outgoing (N).Recorded_State > 0 then
                  --  sent some messages
                  Ada.Text_IO.Put_Line
                    ("         " & Node_ID'Image (N) & ": sent 1 .." &
                     Integer'Image (Outgoing (N).Recorded_State));
               end if;
            end if;
         end loop;
         Ada.Text_IO.Put_Line ("       In channels");
         for N in Node_ID loop
            if Incoming (N).Exists then
               if Incoming (N).Recorded_State > 0 then
                  --  received some messages
                  Ada.Text_IO.Put
                    ("         " & Node_ID'Image (N) &
                     ": received 1 .." &
                     Integer'Image (Incoming (N).Recorded_State));
                  if Incoming (N).Recorded_State /=
                    Incoming (N).Marker_State
                  then
                     Ada.Text_IO.Put_Line
                       (" - stored " &
                        Integer'Image (Incoming (N).Recorded_State + 1) &
                        " .." & Integer'Image (Incoming (N).Marker_State));
                  else
                     Ada.Text_IO.New_Line;
                  end if;
               end if;
            end if;
         end loop;
         Signal (Print); -- unlock
         return True;
      end State_Written;

      task Core_Activity is
         --  a task within task Node to give
         --  each Node individual work
         entry Go;
      end Core_Activity;

      task body Core_Activity is
         Termination : Boolean := False;
      begin
         --  the node waits for the system topology to be deployed
         accept Go;
         --  the node can now start working
         for W in Workload loop
            --  some specific  nodes are interested in snapshotting
            case Current is
               when 1 =>
                  --  node 1 initiates the termination procedure
                  --  at some point in its life span
                  if W = Stop_Point then
                     Termination := True;
                  end if;
               when 2 | 3 | 4 =>
                  null;
            end case;
            if Termination then
               --  node 1 initiates termination
               --  and stops working
               Record_State;
               exit;
            else
               Do_Work (W);
            end if;
         end loop;
         loop
            delay Slow_Down;
            --  this is where Core_Activity should do its own work
            --  until the snapshot has completed
            exit when State_Written;
         end loop;
      exception
         when E : others =>
            Ada.Text_IO.Put_Line (Ada.Exceptions.Exception_Information (E));
      end Core_Activity;

   begin --  Node
      --  configure own topology
      accept Init (ID  : Node_ID;    -- my node ID
                   N_I : Node_Count; -- # inbound channels
                   N_O : Node_Count) -- # outbound channels
      do
         Current  := ID;
         ----------------------
         Last_Wishes.ID := ID;
         ----------------------
         N_In   := N_I;
         N_Out  := N_O;
      end Init;
      ----------------------
      if N_In > 0 then --  configure the inbound channels
         for N in 1 .. N_In loop
            accept Configure (C : Node_ID) do
               Incoming (C).Exists  := True;
            end Configure;
         end loop;
      end if;
      if N_Out > 0 then --  configure the outbound channels
         for N in 1 .. N_Out loop
            accept Configure (C : Node_ID) do
               Outgoing (C).Exists  := True;
            end Configure;
         end loop;
      end if;
      ----------------
      --   this call decouples the job of sending messages out to nodes
      --   from that of receiving them
      --   (hence the activity of every individual node is truly concurrent)
      Core_Activity.Go;
      ----------------
      loop
         select
            accept Message (M : Integer;
                            ID : Node_ID)
            do
               Received_Data  := M;
               Received_ID  := ID;
            end Message;
            if Received_Data < 0 then
               --  a marker has arrived
               if Incoming (Received_ID).Marker_State < 0 then
                  --  record the last msg received on this channel
                  --  after receiving the marker
                  Incoming (Received_ID).Marker_State  :=
                    Incoming (Received_ID).Last_Message;
                  if not State_Recorded then
                     --  start state recording
                     --  if not already done
                     Record_State;
                  end if;
               end if;
            else
               Wait (Sem); -- lock
               Incoming (Received_ID).Last_Message  := Received_Data;
               Signal (Sem); -- unlock
            end if;
         or
            --  gracefully terminates when
            --  there are no more sending nodes
            terminate;
         end select;
      end loop;
   exception
      when E : others =>
         Ada.Text_IO.Put_Line (Ada.Exceptions.Exception_Information (E));
   end Node;

begin -- CL (main)

   --  configure system topology
   Ada.Text_IO.Put_Line
     ("Full node workload runs from" &
      Positive'Image (Workload'First) &
      " to" & Positive'Image (Workload'Last));

   Nodes (1).Init (1, 0, 2);
    --  outgoing
   Nodes (1).Configure (2);
   Nodes (1).Configure (3);

   Nodes (2).Init (2, 2, 1);
    --  incoming
   Nodes (2).Configure (1);
   Nodes (2).Configure (3);
    --  outgoing
   Nodes (2).Configure (4);

   Nodes (3).Init (3, 2, 2);
   --  incoming
   Nodes (3).Configure (1);
   Nodes (3).Configure (4);
   --  outgoing
   Nodes (3).Configure (2);
   Nodes (3).Configure (4);

   Nodes (4).Init (4, 2, 1);
    --  incoming
   Nodes (4).Configure (2);
   Nodes (4).Configure (3);
    --  outgoing
   Nodes (4).Configure (3);

   exception
      when E : others =>
         Ada.Text_IO.Put_Line (Ada.Exceptions.Exception_Information (E));
end CL;
